import logging
import pytest
from ophyd import (Device, Component)

from ophyd import EpicsSignal

logger = logging.getLogger(__name__)


def test_timeout():
    class SubSubDevice(Device):
        cpt5 = Component(EpicsSignal, '5')

    class SubDevice(Device):
        cpt4 = Component(EpicsSignal, '4')
        subsub1 = Component(SubSubDevice, 'sub_')

    class MyDevice(Device):
        sub1 = Component(EpicsSignal, '1')
        sub2 = Component(SubDevice, '_')
        cpt3 = Component(EpicsSignal, '3')

    device = MyDevice('prefix:', name='dev')

    with pytest.raises(TimeoutError) as cm:
        device.wait_for_connection(timeout=1e-6)

    ex_msg = str(cm.value)
    assert 'dev.sub1' in ex_msg
    assert 'dev.sub2.cpt4' in ex_msg
    assert 'dev.sub2.subsub1.cpt5' in ex_msg
    assert 'dev.cpt3' in ex_msg
    assert 'prefix:1' in ex_msg
    assert 'prefix:_4' in ex_msg
    assert 'prefix:_sub_5' in ex_msg
    assert 'prefix:3' in ex_msg



import logging
import pytest
import time
from copy import copy

from ophyd import (PVPositioner, PVPositionerPC, EpicsSignal, EpicsSignalRO,
                   Component as C, get_cl, Kind)

logger = logging.getLogger(__name__)


def setUpModule():
    logging.getLogger('ophyd.pv_positioner').setLevel(logging.DEBUG)
    logger.setLevel(logging.DEBUG)


def tearDownModule():
    logger.debug('Cleaning up')
    logging.getLogger('ophyd.pv_positioner').setLevel(logging.INFO)
    logger.setLevel(logging.INFO)


def test_not_subclassed():
    # can't instantiate it on its own
    with pytest.raises(TypeError):
        PVPositioner('prefix')

    with pytest.raises(TypeError):
        PVPositionerPC('prefix')


def test_no_setpoint_or_readback():
    class MyPositioner(PVPositioner):
        pass

    with pytest.raises(ValueError):
        MyPositioner()


def test_setpoint_but_no_done():
    class MyPositioner(PVPositioner):
        setpoint = C(EpicsSignal, '.VAL')

    with pytest.raises(ValueError):
        MyPositioner()


def test_pvpos(motor):
    class MyPositioner(PVPositioner):
        '''Setpoint, readback, done, stop. No put completion'''
        setpoint = C(EpicsSignal, '.VAL')
        readback = C(EpicsSignalRO, '.RBV')
        done = C(EpicsSignalRO, '.MOVN')
        stop_signal = C(EpicsSignal, '.STOP')

        stop_value = 1
        done_value = 0

    m = MyPositioner(motor.prefix, name='pos_no_put_compl')
    m.wait_for_connection()

    m.read()

    motor.move(0.1, wait=True)
    time.sleep(0.1)
    assert m.position == 0.1

    m.stop()
    m.limits

    repr(m)
    str(m)

    mc = copy(m)
    assert mc.describe() == m.describe()

    m.read()


def test_put_complete_setpoint_only(motor):
    logger.info('--> PV Positioner, using put completion and a DONE pv')

    class MyPositioner(PVPositionerPC):
        '''Setpoint only'''
        setpoint = C(EpicsSignal, '.VAL')

    pos = MyPositioner(motor.prefix, name='pc_setpoint_done')
    print(pos.describe())
    pos.wait_for_connection()

    pos.read()
    high_lim = pos.setpoint.high_limit
    try:
        pos.check_value(high_lim + 1)
    except ValueError as ex:
        logger.info('Check value for single failed, as expected (%s)', ex)
    else:
        raise ValueError('check_value should have failed')

    stat = pos.move(1, wait=False)
    logger.info('--> post-move request, moving=%s', pos.moving)

    while not stat.done:
        logger.info('--> moving... %s error=%s', stat, stat.error)
        time.sleep(0.1)

    pos.move(-1, wait=True)
    assert not pos.moving


def test_put_complete_setpoint_readback_done(motor):
    class MyPositioner(PVPositionerPC):
        '''Setpoint, readback, done, stop. Put completion'''
        setpoint = C(EpicsSignal, '.VAL')
        readback = C(EpicsSignalRO, '.RBV')
        done = C(EpicsSignalRO, '.MOVN')
        done_value = 0

    pos = MyPositioner(motor.prefix, name='pos_no_put_compl',
                       settle_time=0.1, timeout=25.0)
    print(pos.describe())
    pos.wait_for_connection()

    assert pos.settle_time == 0.1
    assert pos.timeout == 25.0
    pos.read()
    high_lim = pos.setpoint.high_limit
    try:
        pos.check_value(high_lim + 1)
    except ValueError as ex:
        logger.info('Check value for single failed, as expected (%s)', ex)
    else:
        raise ValueError('check_value should have failed')

    stat = pos.move(1, wait=False)
    assert stat.timeout == pos.timeout
    logger.info('--> post-move request, moving=%s', pos.moving)

    while not stat.done:
        logger.info('--> moving... %s error=%s', stat, stat.error)
        time.sleep(0.1)

    pos.move(-1, wait=True)
    assert not pos.moving


def test_put_complete_setpoint_readback(motor):
    class MyPositioner(PVPositionerPC):
        '''Setpoint, readback, put completion. No done pv.'''
        setpoint = C(EpicsSignal, '.VAL')
        readback = C(EpicsSignalRO, '.RBV')

    pos = MyPositioner(motor.prefix, name='pos_put_compl')
    print(pos.describe())
    pos.wait_for_connection()

    stat = pos.move(2, wait=False)
    logger.info('--> post-move request, moving=%s', pos.moving)

    while not stat.done:
        logger.info('--> moving... %s', stat)
        time.sleep(0.1)

    pos.move(0, wait=True)
    logger.info('--> synchronous move request, moving=%s', pos.moving)

    time.sleep(0.1)
    print('read', pos.read())
    assert not pos.moving


def test_pvpositioner_with_fake_motor(fake_motor_ioc):
    def callback(sub_type=None, timestamp=None, value=None, **kwargs):
        logger.info('[callback] [%s] (type=%s) value=%s', timestamp,
                    sub_type, value)

    def done_moving(value=0.0, **kwargs):
        logger.info('Done moving %s', kwargs)

    cl = get_cl()
    # ensure we start at 0 for this simple test
    cl.caput(fake_motor_ioc.pvs['setpoint'], 0)
    cl.caput(fake_motor_ioc.pvs['actuate'], 1)
    time.sleep(0.5)

    class MyPositioner(PVPositioner):
        '''Setpoint, readback, no put completion. No done pv.'''
        setpoint = C(EpicsSignal, fake_motor_ioc.pvs['setpoint'])
        readback = C(EpicsSignalRO, fake_motor_ioc.pvs['readback'])
        actuate = C(EpicsSignal, fake_motor_ioc.pvs['actuate'])
        stop_signal = C(EpicsSignal, fake_motor_ioc.pvs['stop'])
        done = C(EpicsSignal, fake_motor_ioc.pvs['moving'])

        actuate_value = 1
        stop_value = 1
        done_value = 0

    pos = MyPositioner('', name='pv_pos_fake_mtr')
    print('fake mtr', pos.describe())
    pos.wait_for_connection()

    pos.subscribe(callback, event_type=pos.SUB_DONE)
    pos.subscribe(callback, event_type=pos.SUB_READBACK)

    logger.info('---- test #1 ----')
    logger.info('--> move to 1')
    pos.move(1, timeout=5)
    assert pos.position == 1
    logger.info('--> move to 0')
    pos.move(0, timeout=5)
    assert pos.position == 0

    logger.info('---- test #2 ----')
    logger.info('--> move to 1')
    pos.move(1, wait=False)
    time.sleep(0.5)
    logger.info('--> stop')
    pos.stop()
    logger.info('--> sleep')
    time.sleep(1)
    logger.info('--> move to 0')
    pos.move(0, wait=False, moved_cb=done_moving)
    logger.info('--> post-move request, moving=%s', pos.moving)
    time.sleep(2)

    pos.read()
    repr(pos)
    str(pos)


def test_hints(fake_motor_ioc):
    class MyPositioner(PVPositioner):
        '''Setpoint, readback, no put completion. No done pv.'''
        setpoint = C(EpicsSignal, fake_motor_ioc.pvs['setpoint'])
        readback = C(EpicsSignalRO, fake_motor_ioc.pvs['readback'])
        actuate = C(EpicsSignal, fake_motor_ioc.pvs['actuate'])
        stop_signal = C(EpicsSignal, fake_motor_ioc.pvs['stop'])
        done = C(EpicsSignal, fake_motor_ioc.pvs['moving'])

        actuate_value = 1
        stop_value = 1
        done_value = 1

    motor = MyPositioner('', name='pv_pos_fake_mtr')

    desc = motor.describe()
    f_hints = motor.hints['fields']
    assert len(f_hints) > 0
    for k in f_hints:
        assert k in desc

    motor.readback.kind = Kind.hinted
    assert motor.hints == {'fields': ['pv_pos_fake_mtr_readback']}

    assert motor.hints['fields'] == f_hints



import time
import logging

import pytest
import operator
import numpy as np


from copy import copy

from ophyd.epics_motor import EpicsMotor
from ophyd.pseudopos import (PseudoPositioner, PseudoSingle)

from ophyd import (Component as C)
from ophyd.utils import ExceptionBundle

logger = logging.getLogger(__name__)


def setUpModule():
    logging.getLogger('ophyd.pseudopos').setLevel(logging.DEBUG)


def tearDownModule():

    logger.debug('Cleaning up')
    logging.getLogger('ophyd.pseudopos').setLevel(logging.INFO)


motor_recs = ['XF:31IDA-OP{Tbl-Ax:X1}Mtr',
              'XF:31IDA-OP{Tbl-Ax:X2}Mtr',
              'XF:31IDA-OP{Tbl-Ax:X3}Mtr',
              'XF:31IDA-OP{Tbl-Ax:X4}Mtr',
              'XF:31IDA-OP{Tbl-Ax:X5}Mtr',
              'XF:31IDA-OP{Tbl-Ax:X6}Mtr',
              ]


class Pseudo3x3(PseudoPositioner):
    pseudo1 = C(PseudoSingle, '', limits=(-10, 10), egu='a')
    pseudo2 = C(PseudoSingle, '', limits=(-10, 10), egu='b')
    pseudo3 = C(PseudoSingle, '', limits=None, egu='c')
    real1 = C(EpicsMotor, motor_recs[0])
    real2 = C(EpicsMotor, motor_recs[1])
    real3 = C(EpicsMotor, motor_recs[2])

    def forward(self, pseudo_pos):
        pseudo_pos = self.PseudoPosition(*pseudo_pos)
        # logger.debug('forward %s', pseudo_pos)
        return self.RealPosition(real1=-pseudo_pos.pseudo1,
                                 real2=-pseudo_pos.pseudo2,
                                 real3=-pseudo_pos.pseudo3)

    def inverse(self, real_pos):
        real_pos = self.RealPosition(*real_pos)
        # logger.debug('inverse %s', real_pos)
        return self.PseudoPosition(pseudo1=real_pos.real1,
                                   pseudo2=real_pos.real2,
                                   pseudo3=real_pos.real3)


class Pseudo1x3(PseudoPositioner):
    pseudo1 = C(PseudoSingle, limits=(-10, 10))
    real1 = C(EpicsMotor, motor_recs[0])
    real2 = C(EpicsMotor, motor_recs[1])
    real3 = C(EpicsMotor, motor_recs[2])

    def forward(self, pseudo_pos):
        pseudo_pos = self.PseudoPosition(*pseudo_pos)
        # logger.debug('forward %s', pseudo_pos)
        return self.RealPosition(real1=-pseudo_pos.pseudo1,
                                 real2=-pseudo_pos.pseudo1,
                                 real3=-pseudo_pos.pseudo1)

    def inverse(self, real_pos):
        real_pos = self.RealPosition(*real_pos)
        # logger.debug('inverse %s', real_pos)
        return self.PseudoPosition(pseudo1=-real_pos.real1)


class FaultyStopperEpicsMotor(EpicsMotor):
    def stop(self, *, success=False):
        raise RuntimeError('Expected exception')


class FaultyPseudo1x3(Pseudo1x3):
    real1 = C(FaultyStopperEpicsMotor, motor_recs[0])


def test_onlypseudo():
    # can't instantiate it on its own
    with pytest.raises(TypeError):
        PseudoPositioner('prefix')


def test_position_wrapper():
    pseudo = Pseudo3x3('', name='mypseudo', concurrent=False)

    test_pos = pseudo.PseudoPosition(pseudo1=1, pseudo2=2, pseudo3=3)
    extra_kw = dict(a=3, b=4, c=6)

    # positional arguments
    assert pseudo.to_pseudo_tuple(1, 2, 3, **extra_kw) == (test_pos, extra_kw)
    # sequence
    assert pseudo.to_pseudo_tuple((1, 2, 3), **extra_kw) == (test_pos, extra_kw)
    # correct type
    assert pseudo.to_pseudo_tuple(test_pos, **extra_kw) == (test_pos, extra_kw)
    # kwargs
    assert pseudo.to_pseudo_tuple(pseudo1=1, pseudo2=2, pseudo3=3, **extra_kw) == \
        (test_pos, extra_kw)

    # too many positional arguments
    with pytest.raises(ValueError):
        pseudo.to_pseudo_tuple(1, 2, 3, 4)
    # valid kwargs, but passing in args too
    with pytest.raises(ValueError):
        pseudo.to_pseudo_tuple(1, pseudo1=1, pseudo2=2, pseudo3=3)


def test_multi_sequential():
    pseudo = Pseudo3x3('', name='mypseudo', concurrent=False)
    pseudo.wait_for_connection()

    assert pseudo.egu == 'a, b, c'

    pos2 = pseudo.PseudoPosition(pseudo1=0, pseudo2=0, pseudo3=0)
    pseudo.set(pos2, wait=True)
    time.sleep(1.0)
    pos1 = pseudo.PseudoPosition(pseudo1=.1, pseudo2=.2, pseudo3=.3)
    pseudo.set(pos1, wait=True)

    pseudo.real1.set(0, wait=True)
    pseudo.real2.set(0, wait=True)
    pseudo.real3.set(0, wait=True)

    pseudo.pseudo1.stop()

    pseudo.real3.set(0, wait=True)


def test_faulty_stopper():
    pseudo = FaultyPseudo1x3('', name='mypseudo', concurrent=False)
    pseudo.wait_for_connection()

    with pytest.raises(ExceptionBundle):
        # smoke-testing for coverage
        pseudo.pseudo1.stop()


def test_limits():
    pseudo = Pseudo3x3('', name='mypseudo', concurrent=True)
    assert pseudo.limits == ((-10, 10), (-10, 10), (0, 0))
    assert pseudo.low_limit == (-10, -10, 0)
    assert pseudo.high_limit == (10, 10, 0)


def test_read_describe():
    pseudo = Pseudo3x3('', name='mypseudo', concurrent=True)
    pseudo.wait_for_connection()
    desc_dict = pseudo.describe()
    desc_keys = ['source', 'upper_ctrl_limit', 'lower_ctrl_limit', 'shape',
                 'dtype', 'units']

    for key in desc_keys:
        assert key in desc_dict['mypseudo_pseudo3']

    read_dict = pseudo.read()
    read_keys = ['value', 'timestamp']
    for key in read_keys:
        assert key in read_dict['mypseudo_pseudo3']

    assert pseudo.read().keys() == pseudo.describe().keys()


def test_multi_concurrent():
    def done(**kwargs):
        logger.debug('** Finished moving (%s)', kwargs)
    pseudo = Pseudo3x3('', name='mypseudo', concurrent=True,
                       settle_time=0.1, timeout=25.0)
    assert pseudo.sequential is False
    assert pseudo.concurrent is True
    assert pseudo.settle_time == 0.1
    assert pseudo.timeout == 25.0
    pseudo.wait_for_connection()
    assert pseudo.connected
    assert tuple(pseudo.pseudo_positioners) == (pseudo.pseudo1, pseudo.pseudo2, pseudo.pseudo3)
    assert tuple(pseudo.real_positioners) == (pseudo.real1, pseudo.real2, pseudo.real3)

    logger.info('Move to (.2, .2, .2), which is (-.2, -.2, -.2) for real '
                'motors')
    pseudo.set(pseudo.PseudoPosition(.2, .2, .2), wait=True)
    logger.info('Position is: %s (moving=%s)', pseudo.position,
                pseudo.moving)
    pseudo.check_value((2, 2, 2))
    pseudo.check_value(pseudo.PseudoPosition(2, 2, 2))
    try:
        pseudo.check_value((2, 2, 2, 3))
    except ValueError as ex:
        logger.info('Check value failed, as expected (%s)', ex)
    real1 = pseudo.real1
    pseudo1 = pseudo.pseudo1
    try:
        pseudo.check_value((real1.high_limit + 1, 2, 2))
    except ValueError as ex:
        logger.info('Check value failed, as expected (%s)', ex)

    ret = pseudo.set((2, 2, 2), wait=False, moved_cb=done)
    assert ret.settle_time == 0.1
    count = 0
    while not ret.done:
        logger.info('Pos=%s %s (err=%s)', pseudo.position, ret, ret.error)
        count += 1
        if count > 1000:
            raise Exception
        time.sleep(0.1)
    logger.info('Single pseudo axis: %s', pseudo1)

    pseudo1.set(0, wait=True, timeout=5)
    assert pseudo1.target == 0
    pseudo1.sync()
    assert pseudo1.target == pseudo1.position
    # coverage
    pseudo1._started_moving

    try:
        pseudo1.check_value(real1.high_limit + 1)
    except ValueError as ex:
        logger.info('Check value for single failed, as expected (%s)', ex)

    logger.info('Move pseudo1 to 0, position=%s', pseudo.position)
    logger.info('pseudo1 = %s', pseudo1.position)

    def single_sub(**kwargs):
        # logger.info('Single sub: %s', kwargs)
        pass

    pseudo1.subscribe(single_sub, pseudo1.SUB_READBACK)
    ret = pseudo1.set(1, wait=False)
    assert pseudo.timeout == ret.timeout
    count = 0
    while not ret.done:
        logger.info('pseudo1.pos=%s Pos=%s %s (err=%s)', pseudo1.position,
                    pseudo.position, ret, ret.error)
        count += 1
        if count > 20:
            raise Exception
        time.sleep(0.1)

    logger.info('pseudo1.pos=%s Pos=%s %s (err=%s)', pseudo1.position,
                pseudo.position, ret, ret.error)
    copy(pseudo)
    pseudo.read()
    pseudo.describe()
    pseudo.read_configuration()
    pseudo.describe_configuration()
    repr(pseudo)
    str(pseudo)
    pseudo.pseudo1.read()
    pseudo.pseudo1.describe()
    pseudo.pseudo1.read_configuration()
    pseudo.pseudo1.describe_configuration()


def test_single_pseudo():
    logger.info('------- Sequential, single pseudo positioner')
    pos = Pseudo1x3('', name='mypseudo', concurrent=False)
    pos.wait_for_connection()
    reals = pos._real

    logger.info('Move to .2, which is (-.2, -.2, -.2) for real motors')
    pos.set((.2, ), wait=True)
    logger.info('Position is: %s (moving=%s)', pos.position, pos.moving)
    logger.info('Real positions: %s', [real.position for real in reals])

    logger.info('Move to -.2, which is (.2, .2, .2) for real motors')
    pos.set((-.2, ), wait=True)
    logger.info('Position is: %s (moving=%s)', pos.position, pos.moving)
    logger.info('Real positions: %s', [real.position for real in reals])

    copy(pos)
    pos.read()
    pos.describe()
    repr(pos)
    str(pos)


@pytest.mark.parametrize(
    'inpargs,inpkwargs,expected_position,expected_kwargs',
    [((1, 2, 3), {}, (1, 2, 3), {}),
     ((1, 2, ), {}, (1, 2, -3), {}),
     ((1, ), {}, (1, -2, -3), {}),

     (((1, 2, 3),), {}, (1, 2, 3), {}),
     (([1, 2, ],), {}, (1, 2, -3), {}),
     (((1, ),), {}, (1, -2, -3), {}),

     ((), {'pseudo1': 1, 'pseudo2': 2, 'pseudo3': 3}, (1, 2, 3), {}),
     ((), {'pseudo1': 1, 'pseudo2': 2}, (1, 2, -3), {}),
     ((), {'pseudo1': 1}, (1, -2, -3), {}),
     ((), {'pseudo1': 1, 'wait': True}, (1, -2, -3), {'wait': True}),

     (({'pseudo1': 1, 'pseudo2': 2, 'pseudo3': 3},), {}, (1, 2, 3), {}),
     (({'pseudo1': 1, 'pseudo2': 2},), {}, (1, 2, -3), {}),
     (({'pseudo1': 1},), {}, (1, -2, -3), {}),
     (({'pseudo1': 1, 'wait': True},), {'timeout': None},
      (1, -2, -3), {'wait': True, 'timeout': None}),


     ((1, 2, 3), {'timeout': 1}, (1, 2, 3), {'timeout': 1}),
     (((1, 2, 3),), {'timeout': 1}, (1, 2, 3), {'timeout': 1}),
     ]
)
def test_pseudo_position_input_3x3(hw, inpargs, inpkwargs,
                                   expected_position, expected_kwargs):
    pseudo3x3 = hw.pseudo3x3
    pseudo3x3.real1.set(1)
    pseudo3x3.real2.set(2)
    pseudo3x3.real3.set(3)

    out, extra_kwargs = pseudo3x3.to_pseudo_tuple(*inpargs, **inpkwargs)
    assert out == pseudo3x3.PseudoPosition(*expected_position)
    assert extra_kwargs == expected_kwargs

    pseudo3x3.set(*inpargs, **inpkwargs)
    assert pseudo3x3.position == pseudo3x3.PseudoPosition(*expected_position)


@pytest.mark.parametrize(
    'inpargs,inpkwargs',
    [((1, 2, 3, 5), {}),
     ((1, 2, 3), {'pseudo1': 1}),
     ((1, 2, 3), {'pseudo2': 1}),
     ((1, ), {'pseudo2': 1,
              'pseudo3': 1}),
     ((1, 2, ), {'pseudo3': 1}),
     ]
)
def test_pseudo_position_fail_3x3(hw, inpargs, inpkwargs):
    pseudo3x3 = hw.pseudo3x3
    with pytest.raises(ValueError):
        pseudo3x3.to_pseudo_tuple(*inpargs, **inpkwargs)


@pytest.mark.parametrize(
    'inpargs,inpkwargs,expected_position,expected_kwargs',
    [((1, 2, 3), {}, (1, 2, 3), {}),
     ((1, 2, ), {}, (1, 2, 3), {}),
     ((1, ), {}, (1, 2, 3), {}),

     (((1, 2, 3),), {}, (1, 2, 3), {}),
     (([1, 2, ],), {}, (1, 2, 3), {}),
     (((1, ),), {}, (1, 2, 3), {}),

     ((), {'real1': 1, 'real2': 2, 'real3': 3}, (1, 2, 3), {}),
     ((), {'real1': 1, 'real2': 2}, (1, 2, 3), {}),
     ((), {'real1': 1}, (1, 2, 3), {}),
     ((), {'real1': 1, 'foo': 'bar'}, (1, 2, 3), {'foo': 'bar'}),

     (({'real1': 1, 'real2': 2, 'real3': 3},), {}, (1, 2, 3), {}),
     (({'real1': 1, 'real2': 2},), {}, (1, 2, 3), {}),
     (({'real1': 1},), {}, (1, 2, 3), {}),
     (({'real1': 1, 'foo': 'bar'},), {'baz': 'buz'},
      (1, 2, 3), {'foo': 'bar', 'baz': 'buz'}),

     ((1, 2, 3), {'foo': 'bar'}, (1, 2, 3), {'foo': 'bar'}),
     ]
)
def test_real_position_input_3x3(hw, inpargs, inpkwargs,
                                 expected_position, expected_kwargs):
    pseudo3x3 = hw.pseudo3x3
    pseudo3x3.real1.set(1)
    pseudo3x3.real2.set(2)
    pseudo3x3.real3.set(3)

    out, extra_kwargs = pseudo3x3.to_real_tuple(*inpargs, **inpkwargs)
    assert out == pseudo3x3.RealPosition(*expected_position)
    assert extra_kwargs == expected_kwargs


@pytest.mark.parametrize(
    'inpargs,inpkwargs',
    [((1, 2, 3, 5), {}),
     ((1, 2, 3), {'real1': 1}),
     ((1, 2, 3), {'real2': 1}),
     ((1, ), {'real2': 1,
              'real3': 1}),
     ((1, 2, ), {'real3': 1}),
     (({'real3': 1, 'foo': 'bar'},), {'foo': 'bizz'}),
     ((), {})
     ]
)
def test_real_position_fail_3x3(hw, inpargs, inpkwargs):
    pseudo3x3 = hw.pseudo3x3
    with pytest.raises(ValueError):
        pseudo3x3.to_real_tuple(*inpargs, **inpkwargs)


def test_single_pseudo_with_sim(hw):
    logger.info('------- Sequential, single pseudo positioner')
    pos = hw.pseudo1x3

    reals = pos._real

    logger.info('Move to .2, which is (-.2, -.2, -.2) for real motors')
    pos.set((.2, ), wait=True)
    logger.info('Position is: %s (moving=%s)', pos.position, pos.moving)
    logger.info('Real positions: %s', [real.position for real in reals])

    logger.info('Move to -.2, which is (.2, .2, .2) for real motors')
    pos.set((-.2, ), wait=True)
    logger.info('Position is: %s (moving=%s)', pos.position, pos.moving)
    logger.info('Real positions: %s', [real.position for real in reals])

    copy(pos)
    pos.read()
    pos.describe()
    repr(pos)
    str(pos)


@pytest.mark.parametrize('typ', ('to_real_tuple', 'to_pseudo_tuple'))
@pytest.mark.parametrize('op', (operator.sub, operator.add))
@pytest.mark.parametrize('a,b', [
    ((0, 0, 0), (1, 1, 1)),
    ((1, 0, 1), (1, 1, 1)),
    ((9, 0, .3), (.3, .1, .5))
])
def test_pseudo_math(hw, a, b, op, typ):
    pos = hw.pseudo3x3
    a, _ = getattr(pos, typ)(a)
    b, _ = getattr(pos, typ)(b)

    # TODO switch to np asserts
    expected = op(np.asarray(a), np.asarray(b))
    assert (np.asarray(op(a, b)) == expected).all()
    assert (np.asarray(op(a, tuple(b))) == expected).all()
    assert (np.asarray(op(a, list(b))) == expected).all()
    assert (np.asarray(op(a, b._asdict())) == expected).all()
    assert (np.asarray(op(a, {})) == a).all()
    assert abs(op(a, b)) == np.sqrt(np.sum(expected ** 2))


def test_pseudo_hints(hw):
    pos = hw.pseudo3x3

    for j in (1, 2, 3):
        p = getattr(pos, 'pseudo{}'.format(j))
        assert p.hints['fields'] == [p.readback.name]
        p.readback.name = 'aardvark{}'.format(j)
        assert p.hints['fields'] == [p.readback.name]

    expected_fields = [getattr(pos, 'pseudo{}'.format(j)).readback.name
                       for j in (1, 2, 3)]
    assert pos.hints['fields'] == expected_fields



import pytest
import logging
from ophyd import Device


logger = logging.getLogger(__name__)


def test_specify_version():
    # Define a versioned Device:
    class MyDevice(Device, version=1, version_type='ioc'):
        ...

    info = MyDevice._class_info_
    assert info == {'version': 1,
                    'versions': {1: MyDevice},
                    'version_type': 'ioc',
                    'version_of': MyDevice,
                    }

    # Define a new version of that Device:
    class MyDevice_V2(MyDevice, version=2, version_of=MyDevice):
        ...

    info = MyDevice_V2._class_info_
    assert info == {'version': 2,
                    'versions': {1: MyDevice,
                                 2: MyDevice_V2},
                    'version_type': 'ioc',
                    'version_of': MyDevice,
                    }

    # Ensure that the original Device has also been updated:
    assert MyDevice._class_info_['versions'] == {1: MyDevice, 2: MyDevice_V2}

    # Define a user device that inherits - but does not define a new version
    class UserDevice(MyDevice_V2):
        ...

    assert UserDevice._class_info_ == {'versions': {1: MyDevice, 2: MyDevice_V2},
                                       'version': 2,
                                       'version_type': 'ioc',
                                       'version_of': MyDevice,
                                       }


def test_version_requires_subclass():
    class MyDevice(Device, version=1):
        ...

    with pytest.raises(RuntimeError):
        class UnrelatedDevice(Device, version=2, version_of=MyDevice):
            ...



# The LogFormatter is adapted light from tornado, which is licensed under
# Apache 2.0. See other_licenses/ in the repository directory.

import logging
import sys
try:
    import colorama
    colorama.init()
except ImportError:
    colorama = None
try:
    import curses
except ImportError:
    curses = None

__all__ = ('set_handler',)


def _stderr_supports_color():
    try:
        if hasattr(sys.stderr, 'isatty') and sys.stderr.isatty():
            if curses:
                curses.setupterm()
                if curses.tigetnum("colors") > 0:
                    return True
            elif colorama:
                if sys.stderr is getattr(colorama.initialise, 'wrapped_stderr',
                                         object()):
                    return True
    except Exception:
        # Very broad exception handling because it's always better to
        # fall back to non-colored logs than to break at startup.
        pass
    return False


class LogFormatter(logging.Formatter):
    """Log formatter used in Tornado, modified for Python3-only ophyd.
    Key features of this formatter are:
    * Color support when logging to a terminal that supports it.
    * Timestamps on every log line.
    * Robust against str/bytes encoding problems.
    This formatter is enabled automatically by
    `tornado.options.parse_command_line` or `tornado.options.parse_config_file`
    (unless ``--logging=none`` is used).
    Color support on Windows versions that do not support ANSI color codes is
    enabled by use of the colorama__ library. Applications that wish to use
    this must first initialize colorama with a call to ``colorama.init``.
    See the colorama documentation for details.
    __ https://pypi.python.org/pypi/colorama
    .. versionchanged:: 4.5
       Added support for ``colorama``. Changed the constructor
       signature to be compatible with `logging.config.dictConfig`.
    """
    DEFAULT_FORMAT = \
        '%(color)s[%(levelname)1.1s %(asctime)s %(module)s:%(lineno)d]%(end_color)s %(message)s'
    DEFAULT_DATE_FORMAT = '%y%m%d %H:%M:%S'
    DEFAULT_COLORS = {
        logging.DEBUG: 4,  # Blue
        logging.INFO: 2,  # Green
        logging.WARNING: 3,  # Yellow
        logging.ERROR: 1,  # Red
    }

    def __init__(self, fmt=DEFAULT_FORMAT, datefmt=DEFAULT_DATE_FORMAT,
                 style='%', color=True, colors=DEFAULT_COLORS):
        r"""
        :arg bool color: Enables color support.
        :arg str fmt: Log message format.
          It will be applied to the attributes dict of log records. The
          text between ``%(color)s`` and ``%(end_color)s`` will be colored
          depending on the level if color support is on.
        :arg dict colors: color mappings from logging level to terminal color
          code
        :arg str datefmt: Datetime format.
          Used for formatting ``(asctime)`` placeholder in ``prefix_fmt``.
        .. versionchanged:: 3.2
           Added ``fmt`` and ``datefmt`` arguments.
        """
        super().__init__(datefmt=datefmt)
        self._fmt = fmt

        self._colors = {}
        if color and _stderr_supports_color():
            if curses is not None:
                # The curses module has some str/bytes confusion in
                # python3.  Until version 3.2.3, most methods return
                # bytes, but only accept strings.  In addition, we want to
                # output these strings with the logging module, which
                # works with unicode strings.  The explicit calls to
                # unicode() below are harmless in python2 but will do the
                # right conversion in python 3.
                fg_color = (curses.tigetstr("setaf") or
                            curses.tigetstr("setf") or "")

                for levelno, code in colors.items():
                    self._colors[levelno] = str(curses.tparm(fg_color, code), "ascii")
                self._normal = str(curses.tigetstr("sgr0"), "ascii")
            else:
                # If curses is not present (currently we'll only get here for
                # colorama on windows), assume hard-coded ANSI color codes.
                for levelno, code in colors.items():
                    self._colors[levelno] = '\033[2;3%dm' % code
                self._normal = '\033[0m'
        else:
            self._normal = ''

    def format(self, record):
        record.message = record.getMessage()
        record.asctime = self.formatTime(record, self.datefmt)

        try:
            record.color = self._colors[record.levelno]
            record.end_color = self._normal
        except KeyError:
            record.color = ''
            record.end_color = ''

        formatted = self._fmt % record.__dict__

        if record.exc_info and not record.exc_text:
            record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            formatted = '{}\n{}'.format(formatted.rstrip(), record.exc_text)
        return formatted.replace("\n", "\n    ")


plain_log_format = "[%(levelname)1.1s %(asctime)s.%(msecs)03d %(module)s:%(lineno)d] %(message)s"
color_log_format = ("%(color)s[%(levelname)1.1s %(asctime)s.%(msecs)03d "
                    "%(module)s:%(lineno)d]%(end_color)s %(message)s")
logger = logging.getLogger('ophyd')


current_handler = None  # overwritten below


def set_handler(file=sys.stdout, datefmt='%H:%M:%S', color=True):
    """
    Set a new handler on the ``logging.getLogger('ophyd')`` logger.

    This function is run at import time with default paramters. If it is run
    again by the user, the handler from the previous invocation is removed (if
    still present) and replaced.

    Parameters
    ----------
    file : object with ``write`` method or filename string
        Default is ``sys.stdout``.
    datefmt : string
        Date format. Default is ``'%H:%M:%S'``.
    color : boolean
        Use ANSI color codes. True by default.

    Returns
    -------
    handler : logging.Handler
        The handler, which has already been added to the 'ophyd' logger.

    Examples
    --------
    Log to a file.

    >>> set_handler(file='/tmp/what_is_happening.txt')

    Include the date along with the time. (The log messages will always include
    microseconds, which are configured separately, not as part of 'datefmt'.)

    >>> set_handler(datefmt="%Y-%m-%d %H:%M:%S")

    Turn off ANSI color codes.

    >>> set_handler(color=False)
    """
    global current_handler
    if isinstance(file, str):
        handler = logging.FileHandler(file)
    else:
        handler = logging.StreamHandler(file)
    if color:
        format = color_log_format
    else:
        format = plain_log_format
    handler.setFormatter(
        LogFormatter(format, datefmt=datefmt))
    if current_handler in logger.handlers:
        logger.removeHandler(current_handler)
    logger.addHandler(handler)
    current_handler = handler
    return handler


# Add a handler with the default parameters at import time.
current_handler = set_handler()



from collections import deque
import time
from threading import RLock
from functools import wraps
from warnings import warn

import logging
import threading
import numpy as np

logger = logging.getLogger(__name__)


class UseNewProperty(RuntimeError):
    ...


# This is used below by StatusBase.
def _locked(func):
    "an decorator for running a method with the instance's lock"
    @wraps(func)
    def f(self, *args, **kwargs):
        with self._lock:
            return func(self, *args, **kwargs)

    return f


class StatusBase:
    """
    This is a base class that provides a single-slot callback for when the
    specific operation has finished.

    Parameters
    ----------
    timeout : float, optional
        The default timeout to use for a blocking wait, and the amount of time
        to wait to mark the operation as failed
    settle_time : float, optional
        The amount of time to wait between the caller specifying that the
        status has completed to running callbacks
    """
    def __init__(self, *, timeout=None, settle_time=None, done=False,
                 success=False):
        super().__init__()
        self._tname = None
        self._lock = RLock()
        self._callbacks = deque()
        self.done = done
        self.success = success
        self.timeout = None

        if settle_time is None:
            settle_time = 0.0

        self.settle_time = float(settle_time)

        if timeout is not None:
            self.timeout = float(timeout)

        if self.done:
            # in the case of a pre-completed status object,
            # don't handle timeout
            return

        if self.timeout is not None and self.timeout > 0.0:
            thread = threading.Thread(target=self._wait_and_cleanup,
                                      daemon=True, name=self._tname)
            self._timeout_thread = thread
            self._timeout_thread.start()

    def _wait_and_cleanup(self):
        """Handle timeout"""
        try:
            if self.timeout is not None:
                timeout = self.timeout + self.settle_time
            else:
                timeout = None
            wait(self, timeout=timeout, poll_rate=0.2)
        except TimeoutError:
            with self._lock:
                if self.done:
                    # Avoid race condition with settling.
                    return
                logger.debug('Status object %s timed out', str(self))
                try:
                    self._handle_failure()
                finally:
                    self._finished(success=False)
        except RuntimeError:
            pass
        finally:
            self._timeout_thread = None

    def _handle_failure(self):
        pass

    def _settled(self):
        """Hook for when status has completed and settled"""
        pass

    def _settle_then_run_callbacks(self, success=True):
        # wait until the settling time is done to mark completion
        if self.settle_time > 0.0:
            time.sleep(self.settle_time)

        with self._lock:
            if self.done:
                # We timed out while waiting for the settle time.
                return
            self.success = success
            self.done = True
            self._settled()

            for cb in self._callbacks:
                cb()
            self._callbacks.clear()

    def _finished(self, success=True, **kwargs):
        """Inform the status object that it is done and if it succeeded

        .. warning::

           kwargs are not used, but are accepted because pyepics gives
           in a bunch of kwargs that we don't care about.  This allows
           the status object to be handed directly to pyepics (but
           this is probably a bad idea for other reason.

           This may be deprecated in the future.

        Parameters
        ----------
        success : bool, optional
           if the action succeeded.
        """
        if self.done:
            return

        if success and self.settle_time > 0:
            # delay gratification until the settle time is up
            self._settle_thread = threading.Thread(
                target=self._settle_then_run_callbacks, daemon=True,
                kwargs=dict(success=success),
            )
            self._settle_thread.start()
        else:
            self._settle_then_run_callbacks(success=success)

    @property
    def callbacks(self):
        """
        Callbacks to be run when the status is marked as finished

        The callback has no arguments ::

            def cb() -> None:

        """
        return self._callbacks

    @property
    @_locked
    def finished_cb(self):
        if len(self.callbacks) == 1:
            warn("The property `finished_cb` is deprecated, and must raise "
                 "an error if a status object has multiple callbacks. Use "
                 "the `callbacks` property instead.", stacklevel=2)
            cb, = self.callbacks
            assert cb is not None
            return cb
        else:
            raise UseNewProperty("The deprecated `finished_cb` property "
                                 "cannot be used for status objects that have "
                                 "multiple callbacks. Use the `callbacks` "
                                 "property instead.")

    @_locked
    def add_callback(self, cb):
        if self.done:
            cb()
        else:
            self._callbacks.append(cb)

    @finished_cb.setter
    @_locked
    def finished_cb(self, cb):
        if not self.callbacks:
            warn("The setter `finished_cb` is deprecated, and must raise "
                 "an error if a status object already has one callback. Use "
                 "the `add_callback` method instead.", stacklevel=2)
            self.add_callback(cb)
        else:
            raise UseNewProperty("The deprecated `finished_cb` setter cannot "
                                 "be used for status objects that already "
                                 "have one callback. Use the `add_callbacks` "
                                 "method instead.")

    def __and__(self, other):
        """
        Returns a new 'composite' status object, AndStatus,
        with the same base API.

        It will finish when both `self` or `other` finish.
        """
        return AndStatus(self, other)


class AndStatus(StatusBase):
    "a Status that has composes two other Status objects using logical and"
    def __init__(self, left, right, **kwargs):
        super().__init__(**kwargs)
        self.left = left
        self.right = right

        def inner():
            with self._lock:
                with self.left._lock:
                    with self.right._lock:
                        l_success = self.left.success
                        r_success = self.right.success
                        l_done = self.left.done
                        r_done = self.right.done

                        # At least one is done.
                        # If it failed, do not wait for the second one.
                        if (not l_success) and l_done:
                            self._finished(success=False)
                        elif (not r_success) and r_done:
                            self._finished(success=False)

                        elif l_success and r_success and l_done and r_done:
                            # Both are done, successfully.
                            self._finished(success=True)
                        # Else one is done, successfully, and we wait for #2,
                        # when this function will be called again.

        self.left.add_callback(inner)
        self.right.add_callback(inner)

    def __repr__(self):
        return "({self.left!r} & {self.right!r})".format(self=self)

    def __str__(self):
        return ('{0}(done={1.done}, '
                'success={1.success})'
                ''.format(self.__class__.__name__, self)
                )


class Status(StatusBase):
    """A basic status object

    Has an optional associated object instance

    Attributes
    ----------
    obj : any or None
        The object
    """
    def __init__(self, obj=None, **kwargs):
        self.obj = obj
        super().__init__(**kwargs)

    def __str__(self):
        return ('{0}(obj={1.obj}, '
                'done={1.done}, '
                'success={1.success})'
                ''.format(self.__class__.__name__, self)
                )

    __repr__ = __str__


class DeviceStatus(StatusBase):
    """Device status

    Parameters
    ----------
    device : obj
    done : bool, optional
        Whether or not the motion has already completed
    success : bool, optional
        If motion has already completed, the status of that motion
    timeout : float, optional
        The default timeout to use for a blocking wait, and the amount of time
        to wait to mark the motion as failed
    settle_time : float, optional
        The amount of time to wait between motion completion and running
        callbacks
    """
    def __init__(self, device, **kwargs):
        self.device = device
        self._watchers = []
        super().__init__(**kwargs)

    def _handle_failure(self):
        super()._handle_failure()
        logger.debug('Trying to stop %s', repr(self.device))
        self.device.stop()

    def __str__(self):
        return ('{0}(device={1.device.name}, done={1.done}, '
                'success={1.success})'
                ''.format(self.__class__.__name__, self)
                )

    def watch(self, func):
        # See MoveStatus.watch for a richer implementation and more info.
        if self.device is not None:
            self._watchers.append(func)
            func(name=self.device.name)

    def _settled(self):
        '''Hook for when status has completed and settled'''
        for watcher in self._watchers:
            watcher(name=self.device.name, fraction=1)

    __repr__ = __str__


class SubscriptionStatus(DeviceStatus):
    """
    Status updated via `ophyd` events

    Parameters
    ----------
    device : obj

    callback : callable
        Callback that takes event information and returns a boolean. Signature
        should be `f(*args, **kwargs)`

    event_type : str, optional
        Name of event type to check whether the device has finished succesfully

    timeout : float, optional
        Maximum timeout to wait to mark the request as a failure

    settle_time : float, optional
        Time to wait after completion until running callbacks

    run: bool, optional
        Run the callback now
    """
    def __init__(self, device, callback, event_type=None,
                 timeout=None, settle_time=None, run=True):
        # Store device and attribute information
        self.device = device
        self.callback = callback

        # Start timeout thread in the background
        super().__init__(device, timeout=timeout, settle_time=settle_time)

        # Subscribe callback and run initial check
        self.device.subscribe(self.check_value,
                              event_type=event_type,
                              run=run)

    def check_value(self, *args, **kwargs):
        """
        Update the status object
        """
        # Get attribute from device
        try:
            success = self.callback(*args, **kwargs)

        # Do not fail silently
        except Exception as e:
            logger.error(e)
            raise

        # If successfull indicate completion
        if success:
            self._finished(success=True)

    def _finished(self, *args, **kwargs):
        """
        Reimplemented finished command to cleanup callback subscription
        """
        # Clear callback
        self.device.clear_sub(self.check_value)
        # Run completion
        super()._finished(**kwargs)


class MoveStatus(DeviceStatus):
    """Asynchronous movement status

    Parameters
    ----------
    positioner : Positioner
    target : float or array-like
        Target position
    done : bool, optional
        Whether or not the motion has already completed
    success : bool, optional
        If motion has already completed, the status of that motion
    start_ts : float, optional
        The motion start timestamp
    timeout : float, optional
        The default timeout to use for a blocking wait, and the amount of time
        to wait to mark the motion as failed
    settle_time : float, optional
        The amount of time to wait between motion completion and running
        callbacks

    Attributes
    ----------
    pos : Positioner
    target : float or array-like
        Target position
    done : bool
        Whether or not the motion has already completed
    start_ts : float
        The motion start timestamp
    finish_ts : float
        The motion completd timestamp
    finish_pos : float or ndarray
        The final position
    success : bool
        Motion successfully completed
    """

    def __init__(self, positioner, target, *, start_ts=None,
                 **kwargs):
        self._tname = 'timeout for {}'.format(positioner.name)
        if start_ts is None:
            start_ts = time.time()

        self.pos = positioner
        self.target = target
        self.start_ts = start_ts
        self.start_pos = self.pos.position
        self.finish_ts = None
        self.finish_pos = None

        self._unit = getattr(self.pos, 'egu', None)
        self._precision = getattr(self.pos, 'precision', None)
        self._name = self.pos.name

        # call the base class
        super().__init__(positioner, **kwargs)

        # Notify watchers (things like progress bars) of new values
        # at the device's natural update rate.
        if not self.done:
            self.pos.subscribe(self._notify_watchers,
                               event_type=self.pos.SUB_READBACK)

    def watch(self, func):
        """
        Subscribe to notifications about progress. Useful for progress bars.

        Parameters
        ----------
        func : callable
            Expected to accept the keyword aruments:

                * ``name``
                * ``current``
                * ``initial``
                * ``target``
                * ``unit``
                * ``precision``
                * ``fraction``
                * ``time_elapsed``
                * ``time_remaining``
        """
        self._watchers.append(func)

    def _notify_watchers(self, value, *args, **kwargs):
        # *args and **kwargs catch extra inputs from pyepics, not needed here
        if not self._watchers:
            return
        current = value
        target = self.target
        initial = self.start_pos
        time_elapsed = time.time() - self.start_ts
        try:
            fraction = abs(target - current) / abs(initial - target)
        # maybe we can't do math?
        except (TypeError, ZeroDivisionError):
            fraction = None

        if np.isnan(fraction):
            fraction = None

        for watcher in self._watchers:
            watcher(name=self._name,
                    current=current,
                    initial=initial,
                    target=target,
                    unit=self._unit,
                    precision=self._precision,
                    time_elapsed=time_elapsed,
                    fraction=fraction)

    @property
    def error(self):
        """Error between target position and current* position

        * If motion is already complete, the final position is used
        """
        if self.finish_pos is not None:
            finish_pos = self.finish_pos
        else:
            finish_pos = self.pos.position

        try:
            return np.array(finish_pos) - np.array(self.target)
        except Exception:
            return None

    def _settled(self):
        """Hook for when motion has completed and settled"""
        super()._settled()
        self.pos.clear_sub(self._notify_watchers)
        self._watchers.clear()
        self.finish_ts = time.time()
        self.finish_pos = self.pos.position

    @property
    def elapsed(self):
        """Elapsed time"""
        if self.finish_ts is None:
            return time.time() - self.start_ts
        else:
            return self.finish_ts - self.start_ts

    def __str__(self):
        return ('{0}(done={1.done}, pos={1.pos.name}, '
                'elapsed={1.elapsed:.1f}, '
                'success={1.success}, settle_time={1.settle_time})'
                ''.format(self.__class__.__name__, self)
                )

    __repr__ = __str__


def wait(status, timeout=None, *, poll_rate=0.05):
    """(Blocking) wait for the status object to complete

    Parameters
    ----------
    timeout : float, optional
        Amount of time in seconds to wait. None disables, such that wait() will
        only return when either the status completes or if interrupted by the
        user.
    poll_rate : float, optional
        Polling rate used to check the status

    Raises
    ------
    TimeoutError
        If time waited exceeds specified timeout
    RuntimeError
        If the status failed to complete successfully
    """
    t0 = time.time()

    def time_exceeded():
        return timeout is not None and (time.time() - t0) > timeout

    while not status.done and not time_exceeded():
        time.sleep(poll_rate)

    if status.done:
        if status.success is not None and not status.success:
            raise RuntimeError('Operation completed but reported an error: {}'
                               ''.format(status))
    elif time_exceeded():
        elapsed = time.time() - t0
        raise TimeoutError('Operation failed to complete within {} seconds '
                           '(elapsed {:.2f} sec)'.format(timeout, elapsed))



''' '''

import logging

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

from .base import *
from .cam import *
from .detectors import *

# NOTE: the following imports are here for backward compatibility with
# previous ophyd versions. This does not represent all available plugins
# in ophyd. For that, import directly from ophyd.areadetector.plugins.
from .plugins import (ColorConvPlugin, FilePlugin, HDF5Plugin, ImagePlugin,
                      JPEGPlugin, MagickPlugin, NetCDFPlugin, NexusPlugin,
                      OverlayPlugin, ProcessPlugin, ROIPlugin, StatsPlugin,
                      TIFFPlugin, TransformPlugin, get_areadetector_plugin,
                      plugin_from_pvname, register_plugin)

from .common_plugins import *
from .trigger_mixins import *



#!/usr/bin/env python
import os
from setuptools import (setup, find_packages)
import versioneer

here = os.path.abspath(os.path.dirname(__file__))

# Get the requirements from requirements.txt
with open(os.path.join(here, 'requirements.txt'), 'rt') as f:
    requirements = f.read().split()

# Get the long description from the README file
with open(os.path.join(here, 'README.md'), 'rt', encoding='utf-8') as f:
    long_description = f.read()

setup(name='ophyd',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      long_description=long_description,
      long_description_content_type='text/markdown',
      license='BSD',
      python_requires='>=3.6',
      install_requires=requirements,
      packages=find_packages(),
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
      ])



import logging
import warnings
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

__all__ = ['setup', 'logger']


def setup():
    from .. import get_cl
    warnings.warn("This function is deprecated in ophyd 1.2 "
                  "and will be removed in ophyd 1.3. "
                  "setup is now automatically called.")
    return get_cl().setup(logger)
